@@ -53,7 +53,7 @@ static int screen_bottom_edge = 0;
 static gint
 expose_handler (GtkTooltips *tooltips)
 {
-  gtk_paint_flat_box (tip->style, tip->window,
+  gtk_paint_flat_box (gtk_widget_get_style (tip), gtk_widget_get_window (tip),
                       GTK_STATE_NORMAL, GTK_SHADOW_OUT, 
                       NULL, tip, "tooltip",
                       0, 0, -1, -1);
